import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

/**
 * A bathroom with a number of toilets and a queue in front of the bathroom.
 * The bathroom is deadlock free because the queue enforces non-circular waiting. With this, the Coffman conditions cannot be satisfied and thus deadlocks cannot occur.
 * The bathroom is starvation free because of the queue in front of the bathroom.
 */
public class Bathroom {
	/**
	 * The queue in front of the toilets.
	 */
	private Queue<Person> queue;
	/**
	 * The occupants of the toilets.
	 */
	private Set<Person> occupants;
	/**
	 * The number of available toilets.
	 */
	private int numberOfToilets;
	
	/**
	 * Create a bathroom with the given number of toilets.
	 * 
	 * @param numberOfToilets	the number of toilets
	 */
	public Bathroom(int numberOfToilets) {
		this.numberOfToilets = numberOfToilets;
		queue = new LinkedList<Person>();
		occupants = new HashSet<Person>();
	}

	/**
	 * Let the entering person get into the bathroom.
	 * 
	 * @param enteringPerson	the entering person
	 */
	public void enter(Person enteringPerson) {
		Person eligiblePerson; /* The eligible person in the queue that can use one of the toilets, or <code>null</code> if none exists. */

		/* Add the entering person to the end of the queue and get the eligible person. */
		synchronized (this) {
			queue.add(enteringPerson);
			eligiblePerson = eligiblePerson();
		}
		
		/* Check whether the person is eligible. */
		if (enteringPerson != eligiblePerson) {
			/* The entering person is not eligible. */
			/* Make the entering person wait in the queue. */
			synchronized (enteringPerson) {
				try {
					enteringPerson.wait();
				} catch (InterruptedException interruptedException) {
					interruptedException.printStackTrace();
				}
			}
		} else {
			/* The entering person is eligible. */
			
			/* Remove the entering person from the queue and assign the entering person to one of the toilets. */
			synchronized (this) {
				queue.remove(enteringPerson);
				occupants.add(enteringPerson);
			}
		}
	}
	
	/**
	 * Make the person leave the bathroom.
	 * 
	 * @param leavingPerson	the person
	 */
	public synchronized void leave(Person leavingPerson) {
		/* Remove the person from the toilets. */
		occupants.remove(leavingPerson);
		
		/* Wake up all the persons that are now eligible to use the toilets due to the leaving person. */
		Person eligiblePerson; /* The eligible person in the queue that can use one of the toilets, or <code>null</code> if none exists. */
		do {
			/* Get the eligible person. */
			eligiblePerson = eligiblePerson();
			
			/* Check whether there is an eligible person. */
			if (eligiblePerson != null) {
				/* There is an eligible person. */
				
				/* Remove the eligible person from the queue and assign the eligible person to one of the toilets. */
				queue.remove(eligiblePerson);
				occupants.add(eligiblePerson);
				
				/* Wake up the eligible person. */
				synchronized (eligiblePerson) {
					eligiblePerson.notify();
				}
			}
		} while (eligiblePerson != null);
	}
	
	/**
	 * Determine the eligible person.
	 * 
	 * @return the person in the queue that is eligible to use on of the toilets, or <code>null</code> if there is no such person in the queue
	 */
	private Person eligiblePerson() {
		Person eligiblePerson; /* The eligible person in the queue that can use one of the toilets, or <code>null</code> if none exists. */
		
		/* Check whether there is someone in the queue and one of the toilets is available. */
		if (queue.size() > 0 && occupants.size() < numberOfToilets) {
			/* Someone is in the queue and one of the toilets is available.*/
			
			/* Check whether there is someone using the toilets. */
			if (occupants.iterator().hasNext()) {
				/* Someone is using the toilets. */
				/* Check which gender is using the toilets. */
				if (occupants.iterator().next() instanceof Man) {
					/* Men are using the toilets. */
					/* Check which gender the first person in the queue has. */
					if (queue.peek() instanceof Man) {
						/* The first person is a man. */
						/* The first person in the queue is eligible. */
						eligiblePerson = queue.peek();
					} else {
						/* The first person is a woman. */
						/* The first person is not eligible. Out of fairness, no one is eligible at this point. */
						eligiblePerson = null;
					}
				} else {
					/* Women are using the toilets. */
					/* Check which gender the first person in the queue has. */
					if (queue.peek() instanceof Woman) {
						/* The first person in the queue is a woman. */
						/* The first person in the queue is eligible. */
						eligiblePerson = queue.peek();
					} else {
						/* The first person in the queue is a man. */
						/* The first person in the queue is not eligible. Out of fairness, no one is eligible at this point. */
						eligiblePerson = null;
					}
				}
			} else {
				/* No one is using the toilets. */
				/* The first person in the queue is eligible. */
				eligiblePerson = queue.peek();
			}
		} else {
			/* Either there is no one in the queue, or there are no toilets available. */
			/* No one is eligible. */
			eligiblePerson = null;
		}
		
		return eligiblePerson;
	}
}
